"""
Main module for the grapheme package.
"""

from .api import graphemes, length, grapheme_lengths, slice, contains, safe_split_index, startswith, endswith

__all__ = ['graphemes', 'length', 'grapheme_lengths', 'slice', 'contains', 'safe_split_index', 'startswith', endswith]
